@@ -20,25 +20,45 @@ try:
 except ImportError:
     MayaDT = None
 
+_TIME_NS_PRESENT = hasattr(time, 'time_ns')
 
 real_time = time.time
 real_localtime = time.localtime
 real_gmtime = time.gmtime
 real_strftime = time.strftime
-real_clock = time.clock
 real_date = datetime.date
 real_datetime = datetime.datetime
 real_date_objects = [real_time, real_localtime, real_gmtime, real_strftime, real_date, real_datetime]
-_real_time_object_ids = set(id(obj) for obj in real_date_objects)
 
-times_to_freeze = []
+if _TIME_NS_PRESENT:
+    real_time_ns = time.time_ns
+    real_date_objects.append(real_time_ns)
+
+_real_time_object_ids = {id(obj) for obj in real_date_objects}
+
+# time.clock is deprecated and was removed in Python 3.8
+real_clock = getattr(time, 'clock', None)
+
+freeze_factories = []
 tz_offsets = []
+ignore_lists = []
+tick_flags = []
+
+# Python3 doesn't have basestring, but it does have str.
+try:
+    # noinspection PyUnresolvedReferences
+    _string_type = basestring
+except NameError:
+    _string_type = str
 
 try:
+    # noinspection PyUnresolvedReferences
     real_uuid_generate_time = uuid._uuid_generate_time
     uuid_generate_time_attr = '_uuid_generate_time'
 except AttributeError:
+    # noinspection PyUnresolvedReferences
     uuid._load_system_functions()
+    # noinspection PyUnresolvedReferences
     real_uuid_generate_time = uuid._generate_time_safe
     uuid_generate_time_attr = '_generate_time_safe'
 except ImportError:
@@ -46,6 +66,7 @@ except ImportError:
     uuid_generate_time_attr = None
 
 try:
+    # noinspection PyUnresolvedReferences
     real_uuid_create = uuid._UuidCreate
 except (AttributeError, ImportError):
     real_uuid_create = None
@@ -57,7 +78,10 @@ except ImportError:
 
 try:
     iscoroutinefunction = inspect.iscoroutinefunction
-    from freezegun._async import wrap_coroutine
+    if sys.version_info < (3, 5):
+        from freezegun._async_coroutine import wrap_coroutine
+    else:
+        from freezegun._async import wrap_coroutine
 except AttributeError:
     iscoroutinefunction = lambda x: False
 
@@ -73,7 +97,7 @@ def _get_module_attributes(module):
     result = []
     try:
         module_attributes = dir(module)
-    except TypeError:
+    except (ImportError, TypeError):
         return result
     for attribute_name in module_attributes:
         try:
@@ -98,9 +122,9 @@ def _setup_module_cache(module):
 def _get_module_attributes_hash(module):
     try:
         module_dir = dir(module)
-    except TypeError:
+    except (ImportError, TypeError):
         module_dir = []
-    return '{0}-{1}'.format(id(module), hash(frozenset(module_dir)))
+    return '{}-{}'.format(id(module), hash(frozenset(module_dir)))
 
 
 def _get_cached_module_attributes(module):
@@ -126,104 +150,90 @@ _is_cpython = (
 )
 
 
-class BaseFakeTime(object):
-    call_stack_inspection_limit = 5
+call_stack_inspection_limit = 5
 
-    def _should_use_real_time(self):
-        if not self.call_stack_inspection_limit:
-            return False
-
-        if not self.ignore:
-            return False
-
-        frame = inspect.currentframe().f_back.f_back
-
-        for _ in range(self.call_stack_inspection_limit):
-            module_name = frame.f_globals.get('__name__')
-            if module_name and module_name.startswith(self.ignore):
-                return True
-
-            frame = frame.f_back
-            if frame is None:
-                break
 
+def _should_use_real_time():
+    if not call_stack_inspection_limit:
         return False
 
+    if not ignore_lists[-1]:
+        return False
 
-class FakeTime(BaseFakeTime):
+    frame = inspect.currentframe().f_back.f_back
 
-    def __init__(self, previous_time_function, ignore=None):
-        self.previous_time_function = previous_time_function
-        self.ignore = ignore
+    for _ in range(call_stack_inspection_limit):
+        module_name = frame.f_globals.get('__name__')
+        if module_name and module_name.startswith(ignore_lists[-1]):
+            return True
 
-    def __call__(self):
-        if self._should_use_real_time():
-            return real_time()
-        current_time = times_to_freeze[-1]()
-        return calendar.timegm(current_time.timetuple()) + current_time.microsecond / 1000000.0
+        frame = frame.f_back
+        if frame is None:
+            break
+
+    return False
 
 
-class FakeLocalTime(BaseFakeTime):
-    def __init__(self, previous_localtime_function=None, ignore=None):
-        self.previous_localtime_function = previous_localtime_function
-        self.ignore = ignore
-
-    def __call__(self, t=None):
-        if t is not None:
-            return real_localtime(t)
-        if self._should_use_real_time():
-            return real_localtime()
-        shifted_time = times_to_freeze[-1]() - datetime.timedelta(seconds=time.timezone)
-        return shifted_time.timetuple()
+def get_current_time():
+    return freeze_factories[-1]()
 
 
-class FakeGMTTime(BaseFakeTime):
-    def __init__(self, previous_gmtime_function, ignore=None):
-        self.previous_gmtime_function = previous_gmtime_function
-        self.ignore = ignore
+def fake_time():
+    if _should_use_real_time():
+        return real_time()
+    current_time = get_current_time()
+    return calendar.timegm(current_time.timetuple()) + current_time.microsecond / 1000000.0
 
-    def __call__(self, t=None):
-        if t is not None:
-            return real_gmtime(t)
-        if self._should_use_real_time():
-            return real_gmtime()
-        return times_to_freeze[-1]().timetuple()
+if _TIME_NS_PRESENT:
+    def fake_time_ns():
+        if _should_use_real_time():
+            return real_time_ns()
+        return int(int(fake_time()) * 1e9)
 
 
-class FakeStrfTime(BaseFakeTime):
-    def __init__(self, previous_strftime_function, ignore=None):
-        self.previous_strftime_function = previous_strftime_function
-        self.ignore = ignore
+def fake_localtime(t=None):
+    if t is not None:
+        return real_localtime(t)
+    if _should_use_real_time():
+        return real_localtime()
+    shifted_time = get_current_time() - datetime.timedelta(seconds=time.timezone)
+    return shifted_time.timetuple()
 
-    def __call__(self, format, time_to_format=None):
-        if time_to_format is None:
-            if not self._should_use_real_time():
-                time_to_format = FakeLocalTime()()
 
+def fake_gmtime(t=None):
+    if t is not None:
+        return real_gmtime(t)
+    if _should_use_real_time():
+        return real_gmtime()
+    return get_current_time().timetuple()
+
+
+def fake_strftime(format, time_to_format=None):
+    if time_to_format is None:
+        if not _should_use_real_time():
+            time_to_format = fake_localtime()
+
+    if time_to_format is None:
+        return real_strftime(format)
+    else:
         return real_strftime(format, time_to_format)
 
-
-class FakeClock(BaseFakeTime):
-    def __init__(self, previous_clock_function, tick=False, ignore=None):
-        self.previous_clock_function = previous_clock_function
-        self.tick = tick
-        self.ignore = ignore
-
-    def __call__(self, *args, **kwargs):
-        if self._should_use_real_time():
+if real_clock is not None:
+    def fake_clock():
+        if _should_use_real_time():
             return real_clock()
 
-        if len(times_to_freeze) == 1:
-            return 0.0 if not self.tick else self.previous_clock_function()
+        if len(freeze_factories) == 1:
+            return 0.0 if not tick_flags[-1] else real_clock()
 
-        first_frozen_time = times_to_freeze[0]()
-        last_frozen_time = times_to_freeze[-1]()
+        first_frozen_time = freeze_factories[0]()
+        last_frozen_time = get_current_time()
 
         timedelta = (last_frozen_time - first_frozen_time)
         total_seconds = timedelta.total_seconds()
 
-        if self.tick:
-            total_seconds += self.previous_clock_function()
+        if tick_flags[-1]:
+            total_seconds += real_clock()
 
         return total_seconds
 
@@ -233,6 +243,10 @@ class FakeDateMeta(type):
     def __instancecheck__(self, obj):
         return isinstance(obj, real_date)
 
+    @classmethod
+    def __subclasscheck__(cls, subclass):
+        return issubclass(subclass, real_date)
+
 
 def datetime_to_fakedatetime(datetime):
     return FakeDatetime(datetime.year,
@@ -277,7 +291,7 @@ class FakeDate(with_metaclass(FakeDateMeta, real_date)):
 
     @staticmethod
     def _date_to_freeze():
-        return times_to_freeze[-1]()
+        return get_current_time()
 
     @classmethod
     def _tz_offset(cls):
@@ -292,6 +306,10 @@ class FakeDatetimeMeta(FakeDateMeta):
     def __instancecheck__(self, obj):
         return isinstance(obj, real_datetime)
 
+    @classmethod
+    def __subclasscheck__(cls, subclass):
+        return issubclass(subclass, real_datetime)
+
 
 class FakeDatetime(with_metaclass(FakeDatetimeMeta, real_datetime, FakeDate)):
     def __new__(cls, *args, **kwargs):
@@ -328,10 +346,11 @@ class FakeDatetime(with_metaclass(FakeDatetimeMeta, real_datetime, FakeDate)):
 
     def date(self):
         return date_to_fakedate(self)
-    
+
     @property
     def nanosecond(self):
         try:
+            # noinspection PyUnresolvedReferences
             return real_datetime.nanosecond
         except AttributeError:
             return 0
@@ -347,8 +366,8 @@ class FakeDatetime(with_metaclass(FakeDatetimeMeta, real_datetime, FakeDate)):
 
     @staticmethod
     def _time_to_freeze():
-        if times_to_freeze:
-            return times_to_freeze[-1]()
+        if freeze_factories:
+            return get_current_time()
 
     @classmethod
     def _tz_offset(cls):
@@ -438,6 +457,7 @@ class FrozenDateTimeFactory(object):
 
     def tick(self, delta=datetime.timedelta(seconds=1)):
         if isinstance(delta, numbers.Real):
+            # noinspection PyTypeChecker
             self.time_to_freeze += datetime.timedelta(seconds=delta)
         else:
             self.time_to_freeze += delta
@@ -449,14 +469,41 @@ class FrozenDateTimeFactory(object):
         self.tick(delta=delta)
 
 
+class StepTickTimeFactory(object):
+
+    def __init__(self, time_to_freeze, step_width):
+        self.time_to_freeze = time_to_freeze
+        self.step_width = step_width
+
+    def __call__(self):
+        return_time = self.time_to_freeze
+        self.tick()
+        return return_time
+
+    def tick(self, delta=None):
+        if not delta:
+            delta = datetime.timedelta(seconds=self.step_width)
+        self.time_to_freeze += delta
+
+    def update_step_width(self, step_width):
+        self.step_width = step_width
+
+    def move_to(self, target_datetime):
+        """Moves frozen date to the given ``target_datetime``"""
+        target_datetime = _parse_time_to_freeze(target_datetime)
+        delta = target_datetime - self.time_to_freeze
+        self.tick(delta=delta)
+
+
 class _freeze_time(object):
 
-    def __init__(self, time_to_freeze_str, tz_offset, ignore, tick, as_arg):
 
+    def __init__(self, time_to_freeze_str, tz_offset, ignore, tick, as_arg, auto_tick_seconds):
         self.time_to_freeze = _parse_time_to_freeze(time_to_freeze_str)
         self.tz_offset = _parse_tz_offset(tz_offset)
         self.ignore = tuple(ignore)
         self.tick = tick
+        self.auto_tick_seconds = auto_tick_seconds
         self.undo_changes = []
         self.modules_at_start = set()
         self.as_arg = as_arg
@@ -477,12 +524,14 @@ class _freeze_time(object):
             orig_setUpClass = getattr(klass, 'setUpClass', None)
             orig_tearDownClass = getattr(klass, 'tearDownClass', None)
 
+            # noinspection PyDecorator
             @classmethod
             def setUpClass(cls):
                 self.start()
                 if orig_setUpClass is not None:
                     orig_setUpClass()
 
+            # noinspection PyDecorator
             @classmethod
             def tearDownClass(cls):
                 if orig_tearDownClass is not None:
@@ -522,33 +571,31 @@ class _freeze_time(object):
         self.stop()
 
     def start(self):
-        if self.tick:
-            time_to_freeze = TickingDateTimeFactory(self.time_to_freeze, real_datetime.now())
-        else:
-            time_to_freeze = FrozenDateTimeFactory(self.time_to_freeze)
 
-        # Change the modules
-        is_already_started = len(times_to_freeze) > 0
-        times_to_freeze.append(time_to_freeze)
+        if self.auto_tick_seconds:
+            freeze_factory = StepTickTimeFactory(self.time_to_freeze, self.auto_tick_seconds)
+        elif self.tick:
+            freeze_factory = TickingDateTimeFactory(self.time_to_freeze, real_datetime.now())
+        else:
+            freeze_factory = FrozenDateTimeFactory(self.time_to_freeze)
+
+        is_already_started = len(freeze_factories) > 0
+        freeze_factories.append(freeze_factory)
         tz_offsets.append(self.tz_offset)
+        ignore_lists.append(self.ignore)
+        tick_flags.append(self.tick)
 
         if is_already_started:
-            return time_to_freeze
+            return freeze_factory
 
+        # Change the modules
         datetime.datetime = FakeDatetime
         datetime.date = FakeDate
 
-        fake_time = FakeTime(time.time, ignore=self.ignore)
-        fake_localtime = FakeLocalTime(time.localtime, ignore=self.ignore)
-        fake_gmtime = FakeGMTTime(time.gmtime, ignore=self.ignore)
-        fake_strftime = FakeStrfTime(time.strftime, ignore=self.ignore)
-        fake_clock = FakeClock(time.clock, tick=self.tick, ignore=self.ignore)
-
         time.time = fake_time
         time.localtime = fake_localtime
         time.gmtime = fake_gmtime
         time.strftime = fake_strftime
-        time.clock = fake_clock
         if uuid_generate_time_attr:
             setattr(uuid, uuid_generate_time_attr, None)
         uuid._UuidCreate = None
@@ -559,17 +606,26 @@ class _freeze_time(object):
 
         # Change any place where the module had already been imported
         to_patch = [
-            ('real_date', real_date, 'FakeDate', FakeDate),
-            ('real_datetime', real_datetime, 'FakeDatetime', FakeDatetime),
-            ('real_gmtime', real_gmtime, 'FakeGMTTime', fake_gmtime),
-            ('real_localtime', real_localtime, 'FakeLocalTime', fake_localtime),
-            ('real_strftime', real_strftime, 'FakeStrfTime', fake_strftime),
-            ('real_time', real_time, 'FakeTime', fake_time),
-            ('real_clock', real_clock, 'FakeClock', fake_clock),
+            ('real_date', real_date, FakeDate),
+            ('real_datetime', real_datetime, FakeDatetime),
+            ('real_gmtime', real_gmtime, fake_gmtime),
+            ('real_localtime', real_localtime, fake_localtime),
+            ('real_strftime', real_strftime, fake_strftime),
+            ('real_time', real_time, fake_time),
         ]
-        self.fake_names = tuple(fake_name for real_name, real, fake_name, fake in to_patch)
-        self.reals = dict((id(fake), real) for real_name, real, fake_name, fake in to_patch)
-        fakes = dict((id(real), fake) for real_name, real, fake_name, fake in to_patch)
+
+        if _TIME_NS_PRESENT:
+            time.time_ns = fake_time_ns
+            to_patch.append(('real_time_ns', real_time_ns, fake_time_ns))
+
+        if real_clock is not None:
+            # time.clock is deprecated and was removed in Python 3.8
+            time.clock = fake_clock
+            to_patch.append(('real_clock', real_clock, fake_clock))
+
+        self.fake_names = tuple(fake.__name__ for real_name, real, fake in to_patch)
+        self.reals = {id(fake): real for real_name, real, fake in to_patch}
+        fakes = {id(real): fake for real_name, real, fake in to_patch}
         add_change = self.undo_changes.append
 
         # Save the current loaded modules
@@ -593,13 +649,15 @@ class _freeze_time(object):
                         setattr(module, attribute_name, fake)
                         add_change((module, attribute_name, attribute_value))
 
-        return time_to_freeze
+        return freeze_factory
 
     def stop(self):
-        times_to_freeze.pop()
+        freeze_factories.pop()
+        ignore_lists.pop()
+        tick_flags.pop()
         tz_offsets.pop()
 
-        if not times_to_freeze:
+        if not freeze_factories:
             datetime.datetime = real_datetime
             datetime.date = real_date
             copyreg.dispatch_table.pop(real_datetime)
@@ -635,11 +693,14 @@ class _freeze_time(object):
                         if real:
                             setattr(module, module_attribute, real)
 
-            time.time = time.time.previous_time_function
-            time.gmtime = time.gmtime.previous_gmtime_function
-            time.localtime = time.localtime.previous_localtime_function
-            time.strftime = time.strftime.previous_strftime_function
-            time.clock = time.clock.previous_clock_function
+            time.time = real_time
+            time.gmtime = real_gmtime
+            time.localtime = real_localtime
+            time.strftime = real_strftime
+            time.clock = real_clock
+
+            if _TIME_NS_PRESENT:
+                time.time_ns = real_time_ns
 
             if uuid_generate_time_attr:
                 setattr(uuid, uuid_generate_time_attr, real_uuid_generate_time)
@@ -666,14 +727,8 @@ class _freeze_time(object):
         return wrapper
 
 
-def freeze_time(time_to_freeze=None, tz_offset=0, ignore=None, tick=False, as_arg=False):
-    # Python3 doesn't have basestring, but it does have str.
-    try:
-        string_type = basestring
-    except NameError:
-        string_type = str
-
-    acceptable_times = (type(None), string_type, datetime.date, datetime.timedelta,
+def freeze_time(time_to_freeze=None, tz_offset=0, ignore=None, tick=False, as_arg=False, auto_tick_seconds=0):
+    acceptable_times = (type(None), _string_type, datetime.date, datetime.timedelta,
              types.FunctionType, types.GeneratorType)
 
     if MayaDT is not None:
@@ -682,15 +737,15 @@ def freeze_time(time_to_freeze=None, tz_offset=0, ignore=None, tick=False, as_ar
     if not isinstance(time_to_freeze, acceptable_times):
         raise TypeError(('freeze_time() expected None, a string, date instance, datetime '
                          'instance, MayaDT, timedelta instance, function or a generator, but got '
-                         'type {0}.').format(type(time_to_freeze)))
+                         'type {}.').format(type(time_to_freeze)))
     if tick and not _is_cpython:
         raise SystemError('Calling freeze_time with tick=True is only compatible with CPython')
 
     if isinstance(time_to_freeze, types.FunctionType):
-        return freeze_time(time_to_freeze(), tz_offset, ignore, tick)
+        return freeze_time(time_to_freeze(), tz_offset, ignore, tick, auto_tick_seconds)
 
     if isinstance(time_to_freeze, types.GeneratorType):
-        return freeze_time(next(time_to_freeze), tz_offset, ignore, tick)
+        return freeze_time(next(time_to_freeze), tz_offset, ignore, tick, auto_tick_seconds)
 
     if MayaDT is not None and isinstance(time_to_freeze, MayaDT):
         return freeze_time(time_to_freeze.datetime(), tz_offset, ignore,
@@ -705,11 +760,16 @@ def freeze_time(time_to_freeze=None, tz_offset=0, ignore=None, tick=False, as_ar
     ignore.append('google.gax')
     ignore.append('threading')
     ignore.append('Queue')
-    return _freeze_time(time_to_freeze, tz_offset, ignore, tick, as_arg)
+    ignore.append('selenium')
+    ignore.append('_pytest.terminal.')
+    ignore.append('_pytest.runner.')
+
+    return _freeze_time(time_to_freeze, tz_offset, ignore, tick, as_arg, auto_tick_seconds)
 
 
 # Setup adapters for sqlite
 try:
+    # noinspection PyUnresolvedReferences
     import sqlite3
 except ImportError:
     # Some systems have trouble with this
